# Copyright (c) 2023-present Bert Van Acker (UA) <bert.vanacker@uantwerpen.be>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import click
from rpio.utils.auxiliary import *


@click.group()
@click.pass_context
def platformCmds():
    pass

@platformCmds.command()
@click.option('--verbose','-v', is_flag=True,default=False,help='Enable debug information.')
@click.option('--check', is_flag=True,default=False,help='Checking the prerequisites for running the adaptive application on this platform.')
@click.option('--set', is_flag=True,default=False,help='Setting up the prerequisites for running the adaptive application on this platform.')
@click.option('--name','-n', default='none', help='Specify the platform name, specified within the AADL.')
@click.option('--force','-f', default='none', help='Force the setup of this platform [native, virtualenv, containerized].')
def platform(verbose,check,set,name,force):
    """Checking the prerequisites for running the adaptive application on this platform."""

    if check:
        if verbose:print("WARNING: platform check is not implemented yet.")

        # CHECK REDIS IS RUNNING
        REDIS_CHECK = check_redis(config=None)
        if REDIS_CHECK:
            print("INFO: REDIS connection check is successful.")
        else:
            print("ERROR: REDIS connection failed. Please check if the platform is connected to the host running the Redis")
            exit()

        #CHECK MQTT IS RUNNING
        MQTT_CHECK = check_mqtt(config=None)
        if MQTT_CHECK:
            print("INFO: MQTT connection check is successful.")
        else:
            print("ERROR: MQTT connection failed. Please check if the platform is connected to the host running the MQTT broker")
            exit()

    if set:

        #NORMAL FLOW, USE AADL INFO FOR SETTING UP THE ENVIRONMENT
        if name != "none":

            # fetch configuration of platform
            platformConfig = "Realization/ManagingSystem/Platform/"+name+"/config.yaml"
            with open(platformConfig, 'r') as file:
                configuration = yaml.safe_load(file)
                formalism = configuration['formalism']
                environmentType = configuration['environment']
                containerization = configuration['containerization']

            # PYTHON-BASED DEPLOYMENT - PYTHON SETUP
            if formalism == "python":

                if type == 'virtualenv':

                    try:
                        create_virtual_environment(venv_name='rpiovenv')
                        launchDescription = parse_launch_xml("Realization/ManagingSystem/Platform/"+name+"/launch.xml")
                        for component in launchDescription.components:
                            install_requirements(venv_name='rpiovenv', requirements_file=component.path+"/requirements.txt")
                        activate_virtual_environment(venv_name='rpiovenv')

                    except:
                        if verbose: print(
                            "ERROR: Could not setup virtual environment for running the adaptive application on this platform.")


            elif formalism == "C++":
                if verbose:print("WARNING: C++ platform setup is not implemented yet.")


        #FORCE FLOW, IGNORING THE AADL INFO
        if force=="virtualenv":
            if verbose: print("INFO: Forcing to setup a virtual python environment for running the adaptive application on this platform.")
        elif force == "native":
            if verbose: print("INFO: Forcing to setup a native python environment for running the adaptive application on this platform.")
        elif force == "containerized":
            if verbose: print("INFO: Forcing to setup containerized environment for running the adaptive application on this platform.")





# **********************************************************************************
# * Copyright (C) 2024-present Bert Van Acker (B.MKR) <bert.vanacker@uantwerpen.be>
# *
# * This file is part of the roboarch R&D project.
# *
# * RAP R&D concepts can not be copied and/or distributed without the express
# * permission of Bert Van Acker
# **********************************************************************************
import tkinter as tk
from tkinter import messagebox
from time import sleep
from threading import Thread
from rpio.utils.constants import *
from rpio.logging.logger import *

class Executer_GUI:

    STATUS_COLORS = {
        StepStatus.PENDING: "#a9a9a9",
        StepStatus.RUNNING: "#1e90ff",
        StepStatus.PASSED: "#32cd32",
        StepStatus.FAILED: "#ff6347"
    }
    STATUS_ICONS = {
        StepStatus.PENDING: "●",  # Grayed-out dot
        StepStatus.RUNNING: "⏳",  # Hourglass for running
        StepStatus.PASSED: "✔️",   # Checkmark for passed
        StepStatus.FAILED: "❌"    # Cross for failed
    }

    def __init__(self, tasks,name):
        self.logger = Logger(name="Custom logger", path="../", verbose=False)
        self.name = name
        self.tasks = tasks
        self.root = tk.Tk()
        self.root.title("Workflow "+name+" Status")
        self.root.geometry("500x400")
        self.root.configure(bg="#1e1e1e")  # Dark background for a polished look
        self.task_widgets = {}
        self.create_gui()

    def create_gui(self):
        title = tk.Label(self.root, text="Workflow "+self.name+" Status", font=("Helvetica", 16, "bold"), bg="#1e1e1e", fg="white")
        title.pack(pady=10)

        self.status_frame = tk.Frame(self.root, bg="#1e1e1e")
        self.status_frame.pack(pady=5, fill="x")

        for task_name in self.tasks:
            self.add_task(task_name)

        self.start_button = tk.Button(
            self.root,
            text="Start Workflow",
            command=self.start_workflow,
            font=("Arial", 12, "bold"),
            bg="#1e90ff",
            fg="white",
            bd=0,
            padx=10,
            pady=5
        )
        self.start_button.pack(pady=20)

    def add_task(self, task_name):
        # Task label with status icon
        task_label = tk.Label(
            self.status_frame,
            text=f"{self.STATUS_ICONS[StepStatus.PENDING]} {task_name}",
            font=("Arial", 12),
            fg=self.STATUS_COLORS[StepStatus.PENDING],
            bg="#1e1e1e",
            anchor="w",
            padx=10,
            pady=5
        )
        task_label.pack(fill="x", pady=2)

        # Store task label for updating later
        self.task_widgets[task_name] = task_label

    def start_workflow(self):
        self.start_button.config(state=tk.DISABLED)
        self.logger.syslog(msg="Workflow --"+self.name+"-- started")
        Thread(target=self.run_tasks).start()

    def run_tasks(self):
        for task_name, func in self.tasks.items():
            self.update_status(task_name, StepStatus.RUNNING)
            try:
                result = func()
                if result:
                    self.update_status(task_name, StepStatus.PASSED)
                    self.logger.syslog(msg="Task <<" + task_name + ">> successfully completed")
                else:
                    self.update_status(task_name, StepStatus.FAILED)
                    self.logger.syslog(msg="Task <<" + task_name + ">> failed to completed")
            except Exception as e:
                self.update_status(task_name, StepStatus.FAILED)
                messagebox.showerror("Error", f"Task '{task_name}' failed with error: {e}")
                break
            sleep(1)  # Simulate delay between checks

        self.start_button.config(state=tk.NORMAL)

    def update_status(self, task_name, status):
        label = self.task_widgets[task_name]
        label.config(
            text=f"{self.STATUS_ICONS[status]} {task_name}",
            fg=self.STATUS_COLORS[status]
        )
        label.update()

class Executer_headless:

    def __init__(self, tasks,name):
        self.tasks = tasks
        self.name = name
        self.logger = Logger(name="Custom logger", path="../", verbose=False)


    def start_workflow(self):
        self.logger.syslog(msg="Workflow " + self.name + " started")
        Thread(target=self.run_tasks).start()

    def run_tasks(self):
        for task_name, func in self.tasks.items():
            try:
                result = func()
                if result:
                    print("Task '{}' succeeded".format(task_name))
                    self.logger.syslog(msg="Task <<" + task_name + ">> successfully completed")
                else:
                    print("Task '{}' failed".format(task_name))
                    self.logger.syslog(msg="Task <<" + task_name + ">> failed to completed")
            except Exception as e:
                print("Error", f"Task '{task_name}' failed with error: {e}")
                break
            sleep(1)  # Simulate delay between checks


# Copyright (c) 2023-present Bert Van Acker (UA) <bert.vanacker@uantwerpen.be>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import click
import os
import subprocess
from rpio.pyLauncher.pyLauncher import launch,launch_main, launch_docker_compose

@click.group()
@click.pass_context
def runCmds():
    pass

@runCmds.command()
@click.option('--verbose','-v', is_flag=True,default=False,help='Enable debug information.')
@click.option('--platform','-p', default='none', help='Specify on which platform you want to run the adaptive application, based on the AADL deployment.')
@click.option('--launchfile', is_flag=True,default=False,help='Specify the use of the launchfile to run the adaptive application.')
@click.option('--docker', is_flag=True,default=False,help='Specify the use of docker to run the adaptive application.')
def run(verbose,platform,launchfile,docker):
    """Run standalone RoboSAPIENS Adaptive Platform application package."""
    if verbose:print("Run command under construction...")

    if platform is None:
        if verbose:print("Executing the run.py action (Realization/ManagingSystem/Actions/run.py)")
        _directory = os.getcwd()
        runFile = "Realization/ManagingSystem/Actions/run.py"
        arguments = ""
        if verbose:print(runFile)

        try:
            subprocess.run(['py.exe',runFile, arguments])
        except:
            print("FAIL - Running standalone robosapiensIO application failed")
    else:
        if docker:
            if verbose: print("Executing the adaptive application using the provided docker compose file for platform {}".format(platform))
            try:
                launch_docker_compose(path='Realization/ManagingSystem/Platform/' + platform)
            except:
                print("FAIL - Launching the standalone robosapiensIO application failed")
        elif launchfile:
            if verbose: print(
                "Executing the adaptive application using the provided launch file for platform {}".format(platform))
            try:
                launch('Realization/ManagingSystem/Platform/' + platform + '/launch.xml')
            except:
                print("FAIL - Launching the standalone robosapiensIO application failed")
        else:
            if verbose:print("Executing the adaptive application using the provided main file for platform {}".format(platform))
            try:
                launch_main('Resources/main_'+platform+'.py')
            except:
                print("FAIL - Launching the standalone robosapiensIO application failed")









from rpio.utils.auxiliary import *
from subprocess import call


def launch(launchFile='launch.xml'):
    """Function to launch one or more python software component using a launch description file

        :param [launchFile]: [launch description file(XML)], defaults to [launch.xml]
        :type [launchFile]: [String](, optional)
        ...
        :return: [Functions returns nothing]
        :rtype: [None]
    """

    # 0. interpret launch file
    launchDescription = parse_launch_xml(file=launchFile,formalism="python")

    # 1. launch all commands at once
    execute_commands(extractCommands(launchDescription))

def launch_main(mainFile='main.py'):
    """Function to launch one or more python software component using a main file

        :param [mainFile]: [launch description file(XML)], defaults to [launch.xml]
        :type [mainFile]: [String](, optional)
        ...
        :return: [Functions returns nothing]
        :rtype: [None]
    """
    command = ["python", mainFile]
    call(command)

def launch_docker_compose(path='/'):
    """Function to launch one or more python software component using a docker compose file

            :param [path]: [path to the docker compose file for the given platform], defaults to [/]
            :type [path]: [String](, optional)
            ...
            :return: [Functions returns nothing]
            :rtype: [None]
        """
    cdCommand = ["cd", path]
    call(cdCommand)
    dockerCommand = ["docker compose up --build -d"]    #-d runs in detached mode
    call(dockerCommand)


# **********************************************************************************
# * Copyright (C) 2024-present Bert Van Acker (B.MKR) <bert.vanacker@uantwerpen.be>
# *
# * This file is part of the roboarch R&D project.
# *
# * RAP R&D concepts can not be copied and/or distributed without the express
# * permission of Bert Van Acker
# **********************************************************************************

from rpio.transformations.transformations import swc2code_py, message2code_py, swc2launch, swc2main, swc2dockerCompose, update_robosapiensIO_ini, add_backbone_config
from rpio.metamodels.aadl2_IL import *
import configparser

def t_load_design():

    # load name and description from ini
    config = configparser.ConfigParser()
    config.read('../robosapiensIO.ini')
    name=config['RoboSAPIENSIO']['name']
    description = config['RoboSAPIENSIO']['description']

    # load design
    design = system(name=name, description=description,JSONDescriptor='../Design/design.json')    #TODO: load AADL when AADL parser is complete

    return design
def t_generate_messages():
    try:
        # load design
        design = t_load_design()
        # generate messages
        message2code_py(system=design, path="../Realization/ManagingSystem/Messages")
        message2code_py(system=design, path="../Realization/ManagedSystem/Messages")
        return True
    except:
        print("Failed to generate the messages")
        return False


def t_generate_swc_skeletons():
    try:
        # load design
        design = t_load_design()
        # generate swc code skeletons
        swc2code_py(system=design, path="../Realization/ManagingSystem/Nodes")
        return True
    except:
        print("Failed to generate the software components")
        return False


def t_generate_swc_launch():
    try:
        # load design
        design = t_load_design()
        # generate launch files
        swc2launch(system=design.systems[0], path="../Realization/ManagingSystem/Platform")
        swc2launch(system=design.systems[1], path="../Realization/ManagedSystem/Platform")
        return True
    except:
        print("Failed to generate the software component launch files")
        return False

def t_generate_main():
    try:
        # load packageName and prefix from ini
        config = configparser.ConfigParser()
        config.read('../robosapiensIO.ini')
        packageName = config['PACKAGE']['name']
        prefix = config['PACKAGE']['prefix']
        # load design
        design = t_load_design()
        # generate main launch file
        if prefix is not "":
            swc2main(system=design.systems[0], package=packageName, prefix=prefix, path="../Resources")
        else:
            swc2main(system=design.systems[0], package=packageName, prefix=None, path="../Resources")
        return True
    except:
        print("Failed to generate the software component main file for the given platforms")
        return False

def t_generate_docker():
    try:
        # load design
        design = t_load_design()
        # generate docker compose file
        swc2dockerCompose(system=design.systems[0], path="../Realization/ManagingSystem/Platform")
        add_backbone_config(system=design, path="../Resources")
        return True
    except:
        print("Failed to generate the docker compose for the given platforms")
        return False

def t_update_robosapiensIO_ini():
    try:
        # load packageName and prefix from ini
        config = configparser.ConfigParser()
        config.read('../robosapiensIO.ini')
        packageName = config['PACKAGE']['name']
        prefix = config['PACKAGE']['prefix']
        # load design
        design = t_load_design()
        # update robosapiensIO ini file
        update_robosapiensIO_ini(system=design,package=packageName,prefix=prefix, path="../")
        return True
    except:
        print("Could not update robosapiensIO.ini")
        return False



import yaml
import logging
from rpio.clientLibraries.rpclpy.communication_manager import CommunicationManager
from rpio.clientLibraries.rpclpy.knowledge import KnowledgeManager

class Node:
    def __init__(self, config, verbose = False):
        self.config = self.load_config(config)
        self.logger = self._initialize_logger()
        self.knowledge = self._initialize_knowledge()  # Initialize knowledge within the component
        self.communication_manager = self._initialize_communication_manager()  # Initialize Event manager
        
        

        # Initialize MQTT and ROS2 Event
        if self.communication_manager:
            self.logger.info(f"{self.__class__.__name__} is using Communication Manager")

    def load_config(self, config_file):
        with open(config_file, 'r') as file:
            return yaml.safe_load(file)

    def _initialize_logger(self):
        """Initialize the logger (same as before)."""
        log_config = self.config.get("logging", {})
        logger = logging.getLogger(self.__class__.__name__)
        log_level = log_config.get("level", "INFO").upper()
        logger.setLevel(getattr(logging, log_level, logging.INFO))

        log_format = log_config.get("format", "%(asctime)s - %(name)s - %(levelname)s - %(message)s")
        log_file = log_config.get("file", None)
        
        formatter = logging.Formatter(log_format)

        if log_file:
            file_handler = logging.FileHandler(log_file)
            file_handler.setFormatter(formatter)
            logger.addHandler(file_handler)
        else:
            console_handler = logging.StreamHandler()
            console_handler.setFormatter(formatter)
            logger.addHandler(console_handler)

        return logger

    def _initialize_knowledge(self):
        """Initialize the Knowledge object based on the config."""
        self.logger.info(f"Initializing Knowledge: {self.config['knowledge_config']['storage_type']} knowledge")
        return KnowledgeManager(self.config['knowledge_config'])

    def _initialize_communication_manager(self):
        """Initialize the Event Manager based on the config."""
        self.logger.info("Initializing Event Manager")
        return CommunicationManager(self.config, self.knowledge, self.logger)

    def start(self):
        """Start the component and enable Event."""
        self.logger.info(f"{self.__class__.__name__} is starting...")
        if self.communication_manager:
            self.communication_manager.start()

    def shutdown(self):
        """Shutdown the component and stop Event."""
        self.logger.info(f"{self.__class__.__name__} is shutting down...")
        if self.communication_manager:
            self.communication_manager.shutdown()

    def publish_event(self, event_key, message = True):
        """Publish Event using the Event manager."""
        if self.communication_manager:
            self.communication_manager.send(event_key, message)
        else:
            self.logger.warning("Event manager is not set for Event publishing.")


    def register_event_callback(self, event_key, callback):
        """Register a callback for Event manager events (MQTT or Redis)."""
        if self.communication_manager:
            self.communication_manager.subscribe(event_key, callback)
            self.logger.info(f"Registered callback for event: {event_key}")
        else:
            self.logger.warning("Event manager is not set for registering event callbacks.")



# **********************************************************************************
# * Copyright (C) 2024-present Bert Van Acker (B.MKR) <bert.vanacker@uantwerpen.be>
# *
# * This file is part of the roboarch R&D project.
# *
# * RAP R&D concepts can not be copied and/or distributed without the express
# * permission of Bert Van Acker
# **********************************************************************************
import os
from pathlib import Path

class PackageManager(object):

    def __init__(self, name='PackageManager',description='Build-in package manager',verbose=False):
        """Initialize a Package Manager component.

                Parameters
                ----------
                name : string
                    name of the property components

                description : string
                    description of the property components

                verbose : bool
                    component verbose execution

        See Also
        --------
        ..

        Examples
        --------
        >> pm = PackageManager(name="PM1",description="Default package manager",verbose=False)

        """

        self._name = name
        self._description = description
        self._verbose = verbose

        #package info
        self._packageName = "rpio_pkg"
        self.standalonePath = ""
        # get current directory
        self._directory = os.getcwd()


    @property
    def name(self):
        """The name property (read-only)."""
        return self._name

    @property
    def description(self):
        """The description property (read-only)."""
        return self._description

    def create(self,name,force=False,standalone=False,path=None):
        """Create rpIO package.

                Parameters
                ----------
                name : string
                    name of the rpIO package

                force : bool
                    force the creation of a rpIO package

                standalone : bool
                    create rpIO package in standalone mode, not in current directory

                path : string
                    path where the new hybridIO package needs to be generated

                See Also
                --------
                ..

                Examples
                --------
                >> pm.create(force=False,standalone=True)

                """
        isEmpty=self._checkEmptyDir()

        if not standalone:
            if isEmpty or (not isEmpty and force):
                if self._verbose:print("DEBUG: directory is empty, creating "+self._packageName+" package...")
                # --- rpio package creation ---
                try:
                    self._populatePackage(name=name, standalone=standalone)
                except:
                    raise Exception("ERROR: "+self._packageName+" package could not be created!")

                if self._verbose: print("DEBUG: "+self._packageName+" package created...")
            else:
                if self._verbose: print("DEBUG: directory is not empty, no "+self._packageName+" package created!")
        else:
            if path is not None:
                self.standalonePath = path
                try:
                    self._populatePackage(name=name, standalone=standalone)
                    if self._verbose: print("DEBUG: " + self._packageName + " package created...")
                except:
                    raise Exception("ERROR: "+self._packageName+" package could not be created!")

            else:
                self.standalonePath = os.getcwd()
                try:
                    self._populatePackage(name=name, standalone=standalone)
                    if self._verbose: print("DEBUG: " + self._packageName + " package created...")
                except:
                    raise Exception("ERROR: "+self._packageName+" package could not be created!")


    def check(self,path=None):
        """Check rpIO package.

            Parameters
            ----------
            path : string
                path to the standalone hybridIO package


            See Also
            --------
            ..

            Examples
            --------
            >> isHybridIOPackage = pm.check(path="path/to/rpio/package")

        """
        validPackage = True
        if path is None:
            if self._verbose: print("DEBUG: checking "+self._packageName+" package...")
            validPackage = os.path.isfile("robosapiensIO.ini")
            #TODO: add other checks
        else:
            if self._verbose: print("DEBUG: checking "+self._packageName+" package in "+self._directory+"/"+path+"...")
            validPackage = os.path.isfile(path+"/robosapiensIO.ini")
            # TODO: add other checks

        return validPackage

    def _checkEmptyDir(self):
        """Function to determine if directory is empty."""
        if self._verbose: print("DEBUG: Checking if directory is empty...")
        dir = os.listdir(self._directory)
        return len(dir) == 0

    def _populatePackage(self, name="rpio_pkg", standalone=False):
        """Function to populate the empty package."""
        self._packageName = name
        if standalone:
            # generate in standalone package instead of in current directory
            Path(self.standalonePath+"/"+self._packageName).mkdir(parents=True, exist_ok=True)
            prefix = self.standalonePath+"/"+self._packageName+'/'
            self._addFile(file="robosapiensIO.ini",name=self._packageName, path=prefix)
            self._addFile(file="__init__.py", name=self._packageName, path=prefix)
            logfilepath = prefix+"/Resources"
        else:
            prefix=""
            self._addFile(file="robosapiensIO.ini",name=self._packageName)
            self._addFile(file="__init__.py", name=self._packageName)
            logfilepath = self._directory + "/Resources"

        self._mkdir_custom(prefix + "Documentation")
        self._mkdir_custom(prefix + "Concept")
        self._mkdir_custom(prefix + "Design")
        self._mkdir_custom(prefix + "Realization")
        #managing system
        self._mkdir_custom(prefix + "Realization/ManagingSystem/Documentation")
        self._mkdir_custom(prefix + "Realization/ManagingSystem/Binaries")
        self._mkdir_custom(prefix + "Realization/ManagingSystem/Nodes")
        self._mkdir_custom(prefix + "Realization/ManagingSystem/Messages")
        self._mkdir_custom(prefix + "Realization/ManagingSystem/Platform")
        self._mkdir_custom(prefix + "Realization/ManagingSystem/Actions")
        self._mkdir_custom(prefix + "Realization/ManagingSystem/Workflows")
        #managed system
        self._mkdir_custom(prefix + "Realization/ManagedSystem/Documentation")
        self._mkdir_custom(prefix + "Realization/ManagedSystem/Binaries")
        self._mkdir_custom(prefix + "Realization/ManagedSystem/Nodes/Probes")
        self._mkdir_custom(prefix + "Realization/ManagedSystem/Nodes/Effector")
        self._mkdir_custom(prefix + "Realization/ManagedSystem/Messages")
        self._mkdir_custom(prefix + "Realization/ManagedSystem/Platform")
        self._mkdir_custom(prefix + "Realization/ManagedSystem/Actions")
        self._mkdir_custom(prefix + "Realization/ManagedSystem/Workflows")
        # top-level workflows
        self._mkdir_custom(prefix + "Workflows")
        self._addFile(file="AADL2CODE.py", path=prefix + "Workflows/")
        # temporary folder and resources
        self._mkdir_custom(prefix + "Resources")

        # add system log file
        self._addFile(file="sys.log", path=logfilepath)

        # add run, build and deploy actions (placeholders) for managing system
        self._addFile(file="run.py", path=prefix +"Realization/ManagingSystem/Actions/")
        self._addFile(file="build.py", path=prefix + "Realization/ManagingSystem/Actions/")
        self._addFile(file="deploy.py", path=prefix + "Realization/ManagingSystem/Actions/")

        # add run, build and deploy actions (placeholders) for managed system
        self._addFile(file="run.py", path=prefix + "Realization/ManagedSystem/Actions/")
        self._addFile(file="build.py", path=prefix + "Realization/ManagedSystem/Actions/")
        self._addFile(file="deploy.py", path=prefix + "Realization/ManagedSystem/Actions/")

    def _mkdir_custom(self, folder="empty", file='readme.md'):
        """CUSTOM mkdir function to initialize git-pushable directories"""
        #create directory
        Path(folder).mkdir(parents=True, exist_ok=True)
        #add file
        self._addFile(file=file, path=folder)

    def _addFile(self, file="requirements.txt",name="",description="", path=None):
        """Function to add a file to the provided path."""

        if path==None:
            path = self._directory

        # --- open file ---
        f = open(path + "/" + file, "a")

        # --- custom file content ---

        if "__init__.py" in file:
            f.write("")

        if "readme" in file:
            f.write("Placeholder")

        if "sys.log" in file:
            f.write("---------------------- RoboSAPIENS Adaptive Platform system logs ----------------------\n")

        if "robosapiensIO.ini" in file:
            f.write("[RoboSAPIENSIO]\n")
            f.write('name = '+name+'\n')
            f.write('description = " Add project description"\n')
            f.write('\n')
            f.write("[PACKAGE]\n")
            f.write("name = "+name+"\n")
            f.write('prefix =  \n')


        if "run.py" in file:
            f.write("print('WARNING: Run action not implemented yet!')")

        if "build.py" in file:
            f.write("print('WARNING: Build action not implemented yet!')")

        if "deploy.py" in file:
            f.write("print('WARNING: Deploy action not implemented yet!')")

        if "AADL2CODE.py" in file:
            f.write("# **********************************************************************************\n")
            f.write("# * Copyright (C) 2024-present Bert Van Acker (B.MKR) <bert.vanacker@uantwerpen.be>\n")
            f.write("# *\n")
            f.write("# * This file is part of the roboarch R&D project.\n")
            f.write("# *\n")
            f.write("# * RAP R&D concepts can not be copied and/or distributed without the express\n")
            f.write("# * permission of Bert Van Acker\n")
            f.write("# **********************************************************************************\n")
            f.write("from rpio.workflow.tasks import *\n")
            f.write("from rpio.workflow.executer import Executer_GUI\n")

            f.write("# 1 . define the tasks\n")
            f.write("tasks = {\n")
            f.write('    "Generate custom messages": t_generate_messages,\n')
            f.write('    "Generate swc code skeletons": t_generate_swc_skeletons,\n')
            f.write('    "Generate swc launch files": t_generate_swc_launch,\n')
            f.write('    "Generate main file": t_generate_main,\n')
            f.write('    "Generate docker compose files": t_generate_docker,\n')
            f.write('    "Update robosapiensIO.ini file": t_update_robosapiensIO_ini\n')
            f.write("}\n")
            f.write("\n")
            f.write("# 2. Launch the graphical executer\n")
            f.write('app = Executer_GUI(tasks=tasks,name="AADL2CODE")\n')
            f.write("app.root.mainloop()\n")

        # --- close file ---
        f.close()





#**************************************************************************
# * Copyright (C) 2023-present Bert Van Acker (B.MKR) <bva.bmkr@gmail.com>
# *
# * This file is part of the hybridIO project.
# *
# * HybridIO can not be copied and/or distributed without the express
# * permission of Bert Van Acker
# *************************************************************************
import os
import logging

def setup_logger(name, log_file, level=logging.INFO):
    """Setup of multiple loggers"""
    formatter = logging.Formatter('%(asctime)s %(levelname)s %(message)s')
    handler = logging.FileHandler(log_file)
    handler.setFormatter(formatter)
    logger = logging.getLogger(name)
    logger.setLevel(level)
    logger.addHandler(handler)

    return logger

class Logger(object):
    def __init__(self,name="customLogger",path=None,verbose=False):
        """Initialize a Logger component.

                Parameters
                ----------
                name : string
                    name of the property components

                verbose : bool
                    component verbose execution

                See Also
                --------
                ..

                Examples
                --------
                >> logger = Logger(verbose=False)

                """

        # --- logger configuration ---
        self._name = name
        self._verbose = verbose

        #set logging file
        if path is None:
            systemLog = os.getcwd()+"/Resources/sys.log"
        else:
            systemLog = path+"/Resources/sys.log"

        #setup logger
        self._syslogger=setup_logger(name="systemLog",log_file=systemLog,level=logging.INFO)
        #self.syslog(msg="Logger configured", level="INFO")

    @property
    def name(self):
        """The name property (read-only)."""
        return self._name

    def syslog(self,msg="tbd",level="INFO"):
        """Perform a system log entry.

            Parameters
            ----------
            msg : string
                message to be logged

            type : string
                Level of the logging


            See Also
            --------
            ..

            Examples
            --------
            >> logger.syslog(msg="this is a examples log message",level="INFO")

            """
        # verbose printing
        if self._verbose: print(msg)
        # logging
        if level == 'INFO':
            self._syslogger.info(msg)
        elif level == 'DEBUG':
            self._syslogger.debug(msg)
        elif level == 'ERROR':
            self._syslogger.error(msg)


# **********************************************************************************
# * Copyright (C) 2024-present Bert Van Acker (B.MKR) <bert.vanacker@uantwerpen.be>
# *
# * This file is part of the roboarch R&D project.
# *
# * RAP R&D concepts can not be copied and/or distributed without the express
# * permission of Bert Van Acker
# **********************************************************************************

import click

from rpio.commands.version import versionCmds
from rpio.commands.importer import importCmds
from rpio.commands.exporter import exportCmds
from rpio.commands.run import runCmds
from rpio.commands.build import buildCmds
from rpio.commands.package import packageCmds
from rpio.commands.deploy import deployCmds
from rpio.commands.transformations import transformationCmds
from rpio.commands.platform import platformCmds

cli=click.CommandCollection(sources=[versionCmds,packageCmds,transformationCmds,runCmds,buildCmds,deployCmds,platformCmds],help="robosapiensIO command line tool")


if __name__ == '__main__':
    cli()



#**********************************************************************************
# * Copyright (C) 2024-present Bert Van Acker (B.MKR) <bert.vanacker@uantwerpen.be>
# *
# * This file is part of the roboarch R&D project.
# *
# * RAP R&D concepts can not be copied and/or distributed without the express
# * permission of Bert Van Acker
# **********************************************************************************
import os
from os import mkdir
from os.path import exists, dirname, join, isfile
import jinja2
from rpio.utils.auxiliary import *
import datetime

#-----------------------------------------------------------------------------------
#-----------------------------------AUXILIARY---------------------------------------
#-----------------------------------------------------------------------------------
def _AddRequirementsFile(file="requirements.txt", path=None):
    """Function to add a requirement.txt to the provided path."""

    # --- open file ---
    f = open(path + "/" + file, "a")

    # --- custom file content ---
    f.write("robosapiensio==0.3.19\n")
    f.write("jsonpickle==3.3.0\n")
    f.write("paho-mqtt==2.1.0\n")
    f.write("PyYAML==6.0.2\n")


    # --- close file ---
    f.close()

def _AddDockerFile(file="Dockerfile",cmpName="", path=None):
    """Function to add a requirement.txt to the provided path."""

    # --- open file ---
    f = open(path + "/" + file, "a")

    # --- custom file content ---
    f.write("FROM python:3.10\n")
    f.write("ENV PYTHONUNBUFFERED 1\n")
    f.write("\n")
    f.write("ADD . c:/src/app\n")
    f.write("WORKDIR c:/src/app\n")
    f.write("ENV PYTHONPATH c:/src/app:$PYTHONPATH\n")
    f.write("\n")
    f.write("COPY requirements.txt ./\n")
    f.write("RUN pip3 install --no-cache-dir -r requirements.txt\n")
    f.write("\n")
    f.write("COPY config.yaml ./\n")
    f.write("COPY messages.py ./\n")
    f.write("COPY . .\n")
    f.write("\n")
    f.write('CMD ["python3", "'+cmpName+'.py"]\n')
    f.write("\n")

    # --- close file ---
    f.close()


#-----------------------------------------------------------------------------------
#--------------------------------TRANSFORMATIONS------------------------------------
#-----------------------------------------------------------------------------------
def swc2code_py(system=None,path="output/generated"):
    """Function to generate python code from the system modeled within the AADL Intermediate Language

    :param [system]: [Adaptive system model within aadlil], defaults to [None]
    :type [system]: [system (aadlil)](, optional)

    :param [path]: [Adaptive system model within aadlil], defaults to ["output/generated/messages"]
    :type [path]: [string](, optional)
    ...
    :return: [Functions returns nothing]
    :rtype: [None]
    """


    if not exists(path):
        mkdir(path)

    # Initialize the Templates engine.
    this_folder = dirname(__file__)
    jinja_env = jinja2.Environment(loader=jinja2.FileSystemLoader(this_folder), trim_blocks=True, lstrip_blocks=True)

    # Load the template
    template = jinja_env.get_template('templates/swc_py.template')
    templateConfig = jinja_env.get_template('templates/swc_config.template')
    templateMessages = jinja_env.get_template('templates/messages_py.template')


    # Extract all processes from AADL system model
    managingSystem = system.systems[0]
    for process in managingSystem.processes:

        # 0. Generate folder for each AADL process
        swcFolder = path + "/" + process.name
        if not exists(swcFolder):
            mkdir(swcFolder)

        #1. Check if SWC file exists
        swcExists = isfile(join(swcFolder, process.name+".py"))

        #2. If swc exists, store custom code added by the user
        if swcExists:
            with open(join(swcFolder, process.name+".py"), "r") as swcFile:
                content = swcFile.read()
                cc_include = getCustomCode(text=content,tag="include")
                cc_code = getCustomCode(text=content, tag="code")
                cc_init = getCustomCode(text=content, tag="init")
                cc_thread_code = []
                for thread in process.threads:
                    _cc_code = getCustomCode(text=content, tag="code_"+thread.name)
                    cc_thread_code.append(_cc_code)

        # 2. Generate code from AADL processes
        with open(join(swcFolder, process.name+".py"), 'w') as f:
            f.write(template.render(swc=process))

        # 3. If swc exists, replace custom code in generated template
        if swcExists:
            with open(join(swcFolder, process.name+".py"), "r") as swcFile:
                content = swcFile.read()
                content = replaceCustomCode(content,tag="include",replacement=cc_include)
                content = replaceCustomCode(content, tag="code", replacement=cc_code)
                content = replaceCustomCode(content, tag="init", replacement=cc_init)
                for i in range(0,len(process.threads),1):
                    content = replaceCustomCode(content, tag="code_"+process.threads[i].name, replacement=cc_thread_code[i])

            with open(join(swcFolder, process.name + ".py"), "w") as swcFile:
                swcFile.write(content)



        # 4. Generate config.yaml file from AADL processes

        #determine the IP address of the platform running the process
        _IP = "localhost"
        for processor in managingSystem.processors:
            if processor.runs_rap_backbone:
                _IP=processor.IP

        with open(join(swcFolder, "config.yaml"), 'w') as f:
            f.write(templateConfig.render(swc=process,IP=_IP))

        # 5. Generate messages for standalone components
        with open(join(swcFolder, "messages.py"), 'w') as f:
            f.write(templateMessages.render(messages=system.messages))

        # 6. Add requirements.txt if swc does not exist
        if not swcExists:
            _AddRequirementsFile(path=swcFolder)

        # 7. Add Docker file if swc does not exist
        if not swcExists:
            _AddDockerFile(cmpName=process.name, path=swcFolder)



def message2code_py(system=None,path="output/generated/messages"):
    """Function to generate python code from messages modeled within the AADL Intermediate Language

    :param [system]: [Adaptive system model within aadlil], defaults to [None]
    :type [system]: [system (aadlil)](, optional)

    :param [path]: [Adaptive system model within aadlil], defaults to ["output/generated/messages"]
    :type [path]: [string](, optional)

    ...
    :return: [Functions returns nothing]
    :rtype: [None]
    """

    if not exists(path):
        mkdir(path)

    # Initialize the Templates engine.
    this_folder = dirname(__file__)
    jinja_env = jinja2.Environment(loader=jinja2.FileSystemLoader(this_folder), trim_blocks=True, lstrip_blocks=True)

    # Load the template
    template = jinja_env.get_template('templates/messages_py.template')

    # Extract all processes from AADL system model
    with open(join(path, "messages.py"), 'w') as f:
        f.write(template.render(messages=system.messages))


def swc2launch(system=None,path="output/generated/lauch"):
    """Function to generate launch files for the given system deployment

    :param [system]: [Managing or managed system model part of the adaptive systen within aadlil,either managing or managed system], defaults to [None]
    :type [system]: [system (aadlil)](, optional)

    :param [path]: [Adaptive system model within aadlil], defaults to ["output/generated/launch"]
    :type [path]: [string](, optional)

    ...
    :return: [Functions returns nothing]
    :rtype: [None]
    """
    if not exists(path):
        mkdir(path)

    # Initialize the Templates engine.
    this_folder = dirname(__file__)
    jinja_env = jinja2.Environment(loader=jinja2.FileSystemLoader(this_folder), trim_blocks=True, lstrip_blocks=True)

    # Load the template
    template = jinja_env.get_template('templates/swc_launch.template')

    # Extract all processors of the managing system
    for processor in system.processors:
        processorPath = join(path, processor.name)
        if not exists(processorPath):
            mkdir(processorPath)

        with open(join(processorPath, "launch.xml"), 'w') as f:
            f.write(template.render(processor=processor))

def swc2main(system=None,package="",prefix=None,path="output/generated/main"):
    """Function to generate main files for the given system deployment

    :param [system]: [Managing or managed system model part of the adaptive systen within aadlil,either managing or managed system], defaults to [None]
    :type [system]: [system (aadlil)](, optional)

    :param [package]: [Package name], defaults to [""]
    :type [package]: [string](, optional)

    :param [path]: [Adaptive system model within aadlil], defaults to ["output/generated/launch"]
    :type [path]: [string](, optional)

    ...
    :return: [Functions returns nothing]
    :rtype: [None]
    """
    if not exists(path):
        mkdir(path)

    # Initialize the Templates engine.
    this_folder = dirname(__file__)
    jinja_env = jinja2.Environment(loader=jinja2.FileSystemLoader(this_folder), trim_blocks=True, lstrip_blocks=True)

    # Load the template
    template = jinja_env.get_template('templates/swc_launch_main.template')

    # Extract all processors of the managing system
    for processor in system.processors:
        with open(join(path, "main_"+processor.name+".py"), 'w') as f:
            f.write(template.render(processor=processor,package=package,prefix=prefix))

def robochart2aadlmessages(maplek=None,path="output/generated/messages"):
    """Function to generate AADL messages from robochart models

    :param [MAPLEK]: [MAPLE-K modeled within robochart], defaults to [None]
    :type [MAPLEK]: [maplek (robochart)](, optional)

    :param [path]: [path to the output folder], defaults to ["output/generated/messages"]
    :type [path]: [string](, optional)

    ...
    :return: [Functions returns nothing]
    :rtype: [None]
    """

    if not exists(path):
        mkdir(path)

    # Initialize the Templates engine.
    this_folder = dirname(__file__)
    jinja_env = jinja2.Environment(loader=jinja2.FileSystemLoader(this_folder), trim_blocks=True, lstrip_blocks=True)

    # Load the template
    template = jinja_env.get_template('templates/aadl_messages.template')

    # Extract all processes from AADL system model
    with open(join(path, "messages.aadl"), 'w') as f:
        f.write(template.render(types=maplek.types))


def swc2dockerCompose(system=None,path="output/generated/docker"):
    """Function to generate docker compose for the given system deployment

    :param [system]: [Managing or managed system model part of the adaptive systen within aadlil,either managing or managed system], defaults to [None]
    :type [system]: [system (aadlil)](, optional)

    :param [path]: [Adaptive system model within aadlil], defaults to ["output/generated/launch"]
    :type [path]: [string](, optional)

    ...
    :return: [Functions returns nothing]
    :rtype: [None]
    """
    if not exists(path):
        mkdir(path)

    # Initialize the Templates engine.
    this_folder = dirname(__file__)
    jinja_env = jinja2.Environment(loader=jinja2.FileSystemLoader(this_folder), trim_blocks=True, lstrip_blocks=True)

    # Load the template
    template = jinja_env.get_template('templates/swc_docker_compose.template')

    # Extract all processors of the managing system
    for processor in system.processors:

        processorPath = join(path, processor.name)
        if not exists(processorPath):
            mkdir(processorPath)

        with open(join(processorPath, "compose.yaml"), 'w') as f:
            f.write(template.render(processor=processor))


def update_robosapiensIO_ini(system=None,package="",prefix ="",path="output/generated/docker"):
    """Function to update the robosapiensIO configuration

    :param [system]: [Managing or managed system model part of the adaptive systen within aadlil,either managing or managed system], defaults to [None]
    :type [system]: [system (aadlil)](, optional)

    :param [path]: [robosapiensIO.ini], defaults to ["output/generated/launch"]
    :type [path]: [string](, optional)

    ...
    :return: [Functions returns nothing]
    :rtype: [None]
    """
    if path is None:
        path = os.getcwd()
    else:
        if not exists(path):
            mkdir(path)

    # Initialize the Templates engine.
    this_folder = dirname(__file__)
    jinja_env = jinja2.Environment(loader=jinja2.FileSystemLoader(this_folder), trim_blocks=True, lstrip_blocks=True)

    # Load the template
    template = jinja_env.get_template('templates/robosapiensIO_ini.template')

    current_timestamp = datetime.datetime.now()
    formatted_timestamp = current_timestamp.strftime('%Y-%m-%d %H:%M:%S')
    managingSystem = system.systems[0]
    managedSystem = system.systems[1]

    with open(join(path, "robosapiensIO.ini"), 'w') as f:
        f.write(template.render(system=system,package=package,prefix=prefix, timestamp=formatted_timestamp.__str__(),managingSystem=managingSystem,managedSystem=managedSystem))


def add_backbone_config(system=None,path='Resources'):
    """Function to add the RoboSAPIENS Adaptive Platform backbone configuration to the repository

    :param [system]: [Managing or managed system model part of the adaptive systen within aadlil,either managing or managed system], defaults to [None]
    :type [system]: [system (aadlil)](, optional)

    :param [path]: [robosapiensIO.ini], defaults to ["output/generated/launch"]
    :type [path]: [string](, optional)

    ...
    :return: [Functions returns nothing]
    :rtype: [None]
    """

    if path is None:
        path = os.getcwd()
    else:
        if not exists(path):
            mkdir(path)

    # Initialize the Templates engine.
    this_folder = dirname(__file__)
    jinja_env = jinja2.Environment(loader=jinja2.FileSystemLoader(this_folder), trim_blocks=True, lstrip_blocks=True)

    # Load the templates
    template_mqtt = jinja_env.get_template('templates/mqtt_config.template')
    template_redis = jinja_env.get_template('templates/redis_config.template')

    with open(join(path, "acl.conf"), 'w') as f:
        f.write(template_mqtt.render(system=system))

    with open(join(path, "redis.conf"), 'w') as f:
        f.write(template_redis.render(system=system))